# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import threading
from functools import wraps


def __gettid():
  """Wrapper for the gettid() system call on Linux systems

  This is a lightweight, fail-fast function to obtain the thread ID of the thread from which it is
  called.  Unfortunately, there's not currently any means of obtaining this information on Linux
  other than through a direct system call. See man 2 gettid() for more details.

  This function can be called directly, but the more common usage pattern is through use of the
  identify_thread decorator on the run() function of a threading.Thread-derived object.

  Returns:
    on success, an int representing the thread ID of the calling thread, as returned from the
      gettid() system call. In the main thread of a process, this should be equal to the process
      ID (e.g. as returned by getpid())
    -1 on any failure (bad platform, error accessing ctypes/libraries, actual system call failure)

  """
  try:
    import platform
    if not platform.system().startswith('Linux'):
      raise ValueError
    syscalls = {
      'i386':   224,   # unistd_32.h: #define __NR_gettid 224
      'x86_64': 186,   # unistd_64.h: #define __NR_gettid 186
    }
    import ctypes
    tid = ctypes.CDLL('libc.so.6').syscall(syscalls[platform.machine()])
  except:
    tid = -1
  return tid


def identify_thread(instancemethod):
  """Simple decorator to expose Linux thread-IDs on an object.

  On Linux, each thread (aka light-weight process) is represented by a unique thread ID, an integer
  in the same namespace as process IDs. (This is distinct from the opaque identifier provided by the
  pthreads interface, which is essentially useless outside the context of the pthreads library.)
  This decorator provides a means to expose this thread ID on Python objects - most likely,
  subclasses of threading.Thread. The benefit of this is that operating-system level process
  information (for example, anything exposed through /proc on Linux) can then be correlated directly
  to Python thread objects.

  The means for retrieving the thread ID is extremely nonportable - specifically, it will only work
  on i386 and x86_64 Linux systems. However, including this decorator more generally should be safe
  and not break any cross-platform code - it will just result in an 'UNKNOWN' thread ID.

  This decorator can be used to wrap any instance method (and technically also class methods). To be
  truly useful, though, it should be used to wrap the run() function of a class utilising the Python
  threading API (i.e. a derivative of threading.Thread)

  Example usage:
    >>> import threading, time
    >>> from twitter.common.decorators import identify_thread
    >>> class MyThread(threading.Thread):
    ...   def __init__(self):
    ...     threading.Thread.__init__(self)
    ...     do_some_other_init()
    ...     self.daemon = True
    ...   @identify_thread
    ...   def run(self):
    ...     while True:
    ...       do_something_awesome()
    ...
    >>> thread1, thread2, thread3 = MyThread(), MyThread(), MyThread()
    >>> thread1.start(), thread2.start(), thread3.start()
    (None, None, None)
    >>> time.sleep(0.1)
    >>> for thread in (thread1, thread2, thread3):
    ...   print '%s => %s' % (thread.name, thread.__thread_id)
    ...
    Thread-1 => 19767
    Thread-2 => 19768
    Thread-3 => 19769
    >>> import os; os.system('ps -L -p %d -o pid,ppid,tid,thcount,cmd' % os.getpid())
      PID  PPID   TID THCNT CMD
    19764 19760 19764     4 /usr/bin/python2.6 /tmp/tmpSW3VIC
    19764 19760 19767     4 /usr/bin/python2.6 /tmp/tmpSW3VIC
    19764 19760 19768     4 /usr/bin/python2.6 /tmp/tmpSW3VIC
    19764 19760 19769     4 /usr/bin/python2.6 /tmp/tmpSW3VIC

  Note that there will be a non-zero delay between when the thread is started and when the thread ID
  attribute (self.__thread_id) is available.

  """
  @wraps(instancemethod)
  def identified(self, *args, **kwargs):
    tid = __gettid()
    if tid == -1:
      self.__thread_id = 'UNKNOWN'
    else:
      self.__thread_id = tid
      if isinstance(self, threading.Thread):
        self.setName('%s [TID=%d]' % (self.name, self.__thread_id))
    return instancemethod(self, *args, **kwargs)

  return identified



# Taken from http://code.activestate.com/recipes/578078
# A 2.6.x backport of Python 3.3 functools.lru_cache
#
# Modified from 4=>2 length indents for code consistency.
#
# Added an on_eviction parameter that gets applied to each element
# when it is evicted from the cache, e.g. file.close(...).

from collections import namedtuple
from functools import update_wrapper
from threading import Lock

_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])

def lru_cache(maxsize=100, typed=False, on_eviction=lambda x:x):
  """Least-recently-used cache decorator.

  If *maxsize* is set to None, the LRU features are disabled and the cache
  can grow without bound.

  If *typed* is True, arguments of different types will be cached separately.
  For example, f(3.0) and f(3) will be treated as distinct calls with
  distinct results.

  Arguments to the cached function must be hashable.

  View the cache statistics named tuple (hits, misses, maxsize, currsize) with
  f.cache_info().  Clear the cache and statistics with f.cache_clear().
  Access the underlying function with f.__wrapped__.

  See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used
  """

  # Users should only access the lru_cache through its public API:
  #       cache_info, cache_clear, and f.__wrapped__
  # The internals of the lru_cache are encapsulated for thread safety and
  # to allow the implementation to change (including a possible C version).
  def decorating_function(user_function):
    cache = dict()
    stats = [0, 0]                  # make statistics updateable non-locally
    HITS, MISSES = 0, 1             # names for the stats fields
    kwd_mark = (object(),)          # separate positional and keyword args
    cache_get = cache.get           # bound method to lookup key or return None
    _len = len                      # localize the global len() function
    lock = Lock()                   # because linkedlist updates aren't threadsafe
    root = []                       # root of the circular doubly linked list
    nonlocal_root = [root]                  # make updateable non-locally
    root[:] = [root, root, None, None]      # initialize by pointing to self
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3    # names for the link fields

    def make_key(args, kwds, typed, tuple=tuple, sorted=sorted, type=type):
      # helper function to build a cache key from positional and keyword args
      key = args
      if kwds:
        sorted_items = tuple(sorted(kwds.items()))
        key += kwd_mark + sorted_items
      if typed:
        key += tuple(type(v) for v in args)
        if kwds:
          key += tuple(type(v) for k, v in sorted_items)
      return key

    if maxsize == 0:
      def wrapper(*args, **kwds):
        # no caching, just do a statistics update after a successful call
        result = user_function(*args, **kwds)
        stats[MISSES] += 1
        return result

    elif maxsize is None:
      def wrapper(*args, **kwds):
        # simple caching without ordering or size limit
        key = make_key(args, kwds, typed) if kwds or typed else args
        result = cache_get(key, root)   # root used here as a unique not-found sentinel
        if result is not root:
          stats[HITS] += 1
          return result
        result = user_function(*args, **kwds)
        cache[key] = result
        stats[MISSES] += 1
        return result

    else:
      def wrapper(*args, **kwds):
        # size limited caching that tracks accesses by recency
        key = make_key(args, kwds, typed) if kwds or typed else args
        with lock:
          link = cache_get(key)
          if link is not None:
            # record recent use of the key by moving it to the front of the list
            root, = nonlocal_root
            link_prev, link_next, key, result = link
            link_prev[NEXT] = link_next
            link_next[PREV] = link_prev
            last = root[PREV]
            last[NEXT] = root[PREV] = link
            link[PREV] = last
            link[NEXT] = root
            stats[HITS] += 1
            return result
        result = user_function(*args, **kwds)
        with lock:
          root = nonlocal_root[0]
          if _len(cache) < maxsize:
            # put result in a new link at the front of the list
            last = root[PREV]
            link = [last, root, key, result]
            cache[key] = last[NEXT] = root[PREV] = link
          else:
            # use root to store the new key and result
            root[KEY] = key
            root[RESULT] = result
            cache[key] = root
            # empty the oldest link and make it the new root
            root = nonlocal_root[0] = root[NEXT]
            evicted = cache.pop(root[KEY])
            on_eviction(evicted[RESULT])
            root[KEY] = None
            root[RESULT] = None
          stats[MISSES] += 1
        return result

    def cache_info():
      """Report cache statistics"""
      with lock:
        return _CacheInfo(stats[HITS], stats[MISSES], maxsize, len(cache))

    def cache_clear():
      """Clear the cache and cache statistics"""
      with lock:
        for value in cache.values():
          on_eviction(value[RESULT])
        cache.clear()
        root = nonlocal_root[0]
        root[:] = [root, root, None, None]
        stats[:] = [0, 0]

    wrapper.__wrapped__ = user_function
    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear
    return update_wrapper(wrapper, user_function)

  return decorating_function
